@@ -31,6 +31,7 @@ import (
 	"sort"
 	"strconv"
 	"strings"
+	"text/template"
 	"time"
 
 	"github.com/docker/go-connections/nat"
@@ -54,6 +55,10 @@ import (
 	goyaml "gopkg.in/yaml.v2"
 )
 
+//go:embed templates/coredns-custom.yaml.tmpl
+var customDNSTemplateStr string
+var customDNSTemplate = template.Must(template.New("customDNS").Parse(customDNSTemplateStr))
+
 // ClusterRun orchestrates the steps of cluster creation, configuration and starting
 func ClusterRun(ctx context.Context, runtime k3drt.Runtime, clusterConfig *config.ClusterConfig) error {
 	/*
@@ -64,13 +69,23 @@ func ClusterRun(ctx context.Context, runtime k3drt.Runtime, clusterConfig *confi
 	}
 
 	// Create tools-node for later steps
-	go EnsureToolsNode(ctx, runtime, &clusterConfig.Cluster)
+
+	g, _ := errgroup.WithContext(ctx)
+	g.Go(func() error {
+		_, err := EnsureToolsNode(ctx, runtime, &clusterConfig.Cluster)
+		return err
+	})
 
 	/*
 	 * Step 1: Create Containers
 	 */
 	if err := ClusterCreate(ctx, runtime, &clusterConfig.Cluster, &clusterConfig.ClusterCreateOpts); err != nil {
-		return fmt.Errorf("Failed Cluster Creation: %+v", err)
+		return fmt.Errorf("failed Cluster Creation: %+v", err)
+	}
+
+	// Wait for tools node to be available
+	if err := g.Wait(); err != nil {
+		return fmt.Errorf("failed to ensure tools node: %w", err)
 	}
 
 	/*
@@ -247,6 +262,25 @@ func ClusterPrep(ctx context.Context, runtime k3drt.Runtime, clusterConfig *conf
 		})
 	}
 
+	/*
+	 * Step 4: Files
+	 */
+
+	for id, node := range clusterConfig.Nodes {
+		for _, nodefile := range node.Files {
+			clusterConfig.Nodes[id].HookActions = append(clusterConfig.Nodes[id].HookActions, k3d.NodeHook{
+				Stage: k3d.LifecycleStagePreStart,
+				Action: actions.WriteFileAction{
+					Runtime:     runtime,
+					Content:     nodefile.Content,
+					Dest:        nodefile.Destination,
+					Mode:        0644,
+					Description: nodefile.Description,
+				},
+			})
+		}
+	}
+
 	return nil
 }
 
@@ -812,7 +846,9 @@ func ClusterGet(ctx context.Context, runtime k3drt.Runtime, cluster *k3d.Cluster
 		for _, existingNode := range cluster.Nodes {
 			// overwrite existing node
 			if existingNode.Name == node.Name {
-				mergo.MergeWithOverwrite(existingNode, node)
+				if err := mergo.Merge(existingNode, node, mergo.WithOverride); err != nil {
+					return nil, fmt.Errorf("failed to merge node %s into cluster: %v", node.Name, err)
+				}
 				overwroteExisting = true
 			}
 		}
@@ -1027,11 +1063,18 @@ func ClusterStart(ctx context.Context, runtime k3drt.Runtime, cluster *k3d.Clust
 			// -> inject hostAliases and network members into CoreDNS configmap
 			if len(servers) > 0 {
 				postStartErrgrp.Go(func() error {
-					hosts := ""
+					type record struct {
+						IP       string
+						Hostname string
+					}
+
+					records := make([]record, 0)
 
 					// hosts: hostAliases (including host.k3d.internal)
 					for _, hostAlias := range clusterStartOpts.HostAliases {
-						hosts += fmt.Sprintf("%s %s\n", hostAlias.IP, strings.Join(hostAlias.Hostnames, " "))
+						for _, hostname := range hostAlias.Hostnames {
+							records = append(records, record{IP: hostAlias.IP, Hostname: hostname})
+						}
 					}
 
 					// more hosts: network members ("neighbor" containers)
@@ -1040,45 +1083,21 @@ func ClusterStart(ctx context.Context, runtime k3drt.Runtime, cluster *k3d.Clust
 						return fmt.Errorf("failed to get cluster network %s to inject host records into CoreDNS: %w", cluster.Network.Name, err)
 					}
 					for _, member := range net.Members {
-						hosts += fmt.Sprintf("%s %s\n", member.IP.String(), member.Name)
+						records = append(records, record{IP: member.IP.String(), Hostname: member.Name})
 					}
 
 					// inject CoreDNS configmap
 					l.Log().Infof("Injecting records for hostAliases (incl. host.k3d.internal) and for %d network members into CoreDNS configmap...", len(net.Members))
-					act := actions.RewriteFileAction{
+					var custom_dns bytes.Buffer
+					err = customDNSTemplate.Execute(&custom_dns, records)
+					if err != nil {
+						return fmt.Errorf("failed to render template: %w", err)
+					}
+					act := actions.WriteFileAction{
 						Runtime: runtime,
-						Path:    "/var/lib/rancher/k3s/server/manifests/coredns.yaml",
+						Content: []byte(custom_dns.Bytes()),
+						Dest:    "/var/lib/rancher/k3s/server/manifests/coredns-custom.yaml",
 						Mode:    0744,
-						RewriteFunc: func(input []byte) ([]byte, error) {
-							split, err := util.SplitYAML(input)
-							if err != nil {
-								return nil, fmt.Errorf("error splitting yaml: %w", err)
-							}
-
-							var outputBuf bytes.Buffer
-							outputEncoder := util.NewYAMLEncoder(&outputBuf)
-
-							for _, d := range split {
-								var doc map[string]interface{}
-								if err := yaml.Unmarshal(d, &doc); err != nil {
-									return nil, err
-								}
-								if kind, ok := doc["kind"]; ok {
-									if strings.ToLower(kind.(string)) == "configmap" {
-										configmapData, ok := doc["data"].(map[string]interface{})
-										if !ok {
-											return nil, fmt.Errorf("invalid ConfigMap data type: %T", doc["data"])
-										}
-										configmapData["NodeHosts"] = hosts
-									}
-								}
-								if err := outputEncoder.Encode(doc); err != nil {
-									return nil, err
-								}
-							}
-							_ = outputEncoder.Close()
-							return outputBuf.Bytes(), nil
-						},
 					}
 
 					// get the first server in the list and run action on it once it's ready for it
@@ -1097,7 +1116,7 @@ func ClusterStart(ctx context.Context, runtime k3drt.Runtime, cluster *k3d.Clust
 						if err := NodeWaitForLogMessage(postStartErrgrpCtx, runtime, n, "Cluster dns configmap", ts.Truncate(time.Second)); err != nil {
 							return err
 						}
-						return act.Run(postStartErrgrpCtx, n)
+						return act.Run(postStartErrgrpCtx, n) // nolint:staticcheck // FIXME: Does this loop really only concern the first server? (SA4004: the surrounding loop is unconditionally terminated (staticcheck))
 					}
 					return nil
 				})
@@ -1234,7 +1253,9 @@ func ClusterEditChangesetSimple(ctx context.Context, runtime k3drt.Runtime, clus
 	}
 	lbChangeset.Node.HookActions = append(lbChangeset.Node.HookActions, writeLbConfigAction)
 
-	NodeReplace(ctx, runtime, existingLB.Node, lbChangeset.Node)
+	if err := NodeReplace(ctx, runtime, existingLB.Node, lbChangeset.Node); err != nil {
+		return fmt.Errorf("error replacing loadbalancer node: %w", err)
+	}
 
 	return nil
 }
