@@ -21,12 +21,12 @@ export KUBECONFIG
 : "${EXTRA_FLAG:=""}"
 : "${EXTRA_TITLE:=""}"
 
-if [[ -n "$K3S_IMAGE_TAG" ]]; then
-  EXTRA_FLAG="--image rancher/k3s:$K3S_IMAGE_TAG"
-  EXTRA_TITLE="(rancher/k3s:$K3S_IMAGE_TAG)"
+if [[ -n "$K3S_IMAGE" ]]; then
+    EXTRA_FLAG="--image $K3S_IMAGE"
+    EXTRA_TITLE="($K3S_IMAGE)"
 fi
 
-export CURRENT_STAGE="Test | lifecycle | $K3S_IMAGE_TAG"
+export CURRENT_STAGE="Test | lifecycle | $K3S_IMAGE"
 
 
 clustername="lifecycletest"
@@ -60,7 +60,6 @@ $EXE cluster start "$clustername" --wait --timeout 360s || failed "cluster didn'
 info "Checking that we have access to the cluster..."
 check_clusters "$clustername" || failed "error checking cluster"
 
-kubectl delete pod -n kube-system -l k8s-app=kube-dns  > /dev/null 2>&1 # delete coredns to force reload of config (reload plugin uses default 30s, which will make tests below fail)
 
 info "Checking that we have 2 nodes online..."
 check_multi_node "$clustername" 2 || failed "failed to verify number of nodes"
@@ -74,25 +73,26 @@ check_multi_node "$clustername" 3 || failed "failed to verify number of nodes"
 
 # 4. load an image into the cluster
 info "Importing an image into the cluster..."
-docker pull alpine:latest > /dev/null
-docker tag alpine:latest alpine:local > /dev/null
-$EXE image import alpine:local -c $clustername || failed "could not import image in $clustername"
+docker pull iwilltry42/dnsutils:20.04 > /dev/null
+docker tag iwilltry42/dnsutils:20.04 testimage:local > /dev/null
+$EXE image import testimage:local -c $clustername || failed "could not import image in $clustername"
 
 # 5. use imported image
 info "Spawning a pod using the imported image..."
-kubectl run --image alpine:local testimage --command -- tail -f /dev/null
+kubectl run --image testimage:local testimage --command -- tail -f /dev/null
 info "Waiting for a bit for the pod to start..."
 sleep 5
 
+kubectl delete pod -n kube-system -l k8s-app=kube-dns  > /dev/null 2>&1 # delete coredns to force reload of config (reload plugin uses default 30s, which will make tests below fail)
 wait_for_pod_running_by_name "testimage"
 wait_for_pod_running_by_label "k8s-app=kube-dns" "kube-system"
 
 sleep 5
 
 # 6. test host.k3d.internal
-info "Checking DNS Lookup for host.k3d.internal..."
-kubectl describe cm coredns -n kube-system | grep "host.k3d.internal" > /dev/null 2>&1 || failed "Couldn't find host.k3d.internal in CoreDNS configmap"
-wait_for_pod_exec "testimage" "nslookup host.k3d.internal" 15 || failed "DNS Lookup for host.k3d.internal failed"
+info "Checking DNS Lookup for host.k3d.internal via Ping..."
+kubectl describe cm coredns-custom -n kube-system | grep "host.k3d.internal" > /dev/null 2>&1 || failed "Couldn't find host.k3d.internal in coredns-custom configmap"
+wait_for_pod_exec "testimage" "ping -c1 host.k3d.internal" 15 || failed "Pinging host.k3d.internal failed"
 
 # Cleanup
 
